#include-once
#include "Bass.au3"
#include "BassMidiConstants.au3"

Global $_ghBassMidiDll = -1
Global $BASS_MIDI_DLL_UDF_VER = "2.4.6.3"
Global $BASS_MIDI_UDF_VER = "10.0"

; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_Startup
; Description ...: Starts up BASSMIDI functions.
; Syntax ........: _BASS_MIDI_Startup($sBassMidiDLL = "")
; Parameters ....: -	$sBassMidiDLL	-	The relative path to BassMidi.dll.
; Return values .: Success      - Returns True
;                  Failure      - Returns False and sets @ERROR
;									@error will be set to-
;										- $BASS_ERR_DLL_NO_EXIST	-	File could not be found.
; Author ........: Prog@ndy
; Modified ......: Eukalyptus
; Remarks .......:
; Related .......:
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_Startup($sBassMidiDLL = "")
	If $_ghBassMidiDll <> -1 Then Return True
	If Not $sBassMidiDLL Then $sBassMidiDLL = @ScriptDir & "\BassMidi.dll"

	If Not FileExists($sBassMidiDLL) Then Return SetError($BASS_ERR_DLL_NO_EXIST, 0, False)

	Local $sBit = __BASS_LibraryGetArch($sBassMidiDLL)
	Select
		Case $sBit = "32" And @AutoItX64
			ConsoleWrite(@CRLF & "!BassMidi.dll is for 32bit only!" & @CRLF & "Run/compile Script at 32bit" & @CRLF)
		Case $sBit = "64" And Not @AutoItX64
			ConsoleWrite(@CRLF & "!BassMidi.dll is for 64bit only!" & @CRLF & "use 32bit version of BassMidi.dll" & @CRLF)
	EndSelect

	If $BASS_STARTUP_VERSIONCHECK Then
		If Not @AutoItX64 And _VersionCompare(FileGetVersion($sBassMidiDLL), $BASS_MIDI_DLL_UDF_VER) <> 0 Then ConsoleWrite(@CRLF & "!This version of BASSMIDI.au3 is made for BassMIDI.dll V" & $BASS_MIDI_DLL_UDF_VER & ".  Please update" & @CRLF)
		If $BASS_MIDI_UDF_VER <> $BASS_UDF_VER Then ConsoleWrite("!This version of BASSMIDI.au3 (v" & $BASS_MIDI_UDF_VER & ") may not be compatible to BASS.au3 (v" & $BASS_UDF_VER & ")" & @CRLF)
	EndIf

	$_ghBassMidiDll = DllOpen($sBassMidiDLL)
	Return $_ghBassMidiDll <> -1
EndFunc   ;==>_BASS_MIDI_Startup



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontCompact
; Description ...: Compacts a soundfont's memory usage.
; Syntax ........: _BASS_MIDI_FontCompact($handle)
; Parameters ....: - $handle - The soundfont handle... 0 = all soundfonts.
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Compacting involves freeing any samples that are currently loaded but unused. The amount of sample data currently loaded can be retrieved using BASS_MIDI_FontGetInfo.
; Related .......: BASS_MIDI_FontFree, BASS_MIDI_FontGetInfo, BASS_CONFIG_MIDI_COMPACT
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontCompact($handle)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontCompact", "DWORD", $handle)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontCompact



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontFree
; Description ...: Frees a soundfont.
; Syntax ........: _BASS_MIDI_FontFree($handle)
; Parameters ....: - $handle - The soundfont handle.
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: When a soundfont is freed, it is automatically removed from any MIDI streams that are using it.
; Related .......: BASS_MIDI_FontCompact, BASS_MIDI_FontInit
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontFree($handle)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontFree", "DWORD", $handle)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontFree



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontGetInfo
; Description ...: Retrieves information on a soundfont.
; Syntax ........: _BASS_MIDI_FontGetInfo($handle)
; Parameters ....: - $handle - The soundfont to get info on.
; Return values .: Success - Returns an array containg information on a soundfont.
;								- [0] = name Name of the soundfont.
;								- [1] = copyright Copyright notice.
;								- [2] = comment Any comments.
;								- [3] = presets The number of presets/instruments in the soundfont.
;								- [4] = samsize The total size (in bytes) of the sample data in the soundfont.
;								- [5] = samload The amount of sample data currently loaded.
;								- [6] = samtype The BASS_CTYPE_STREAM_xxx format of the sample data if it's packed... -1 = unknown format (appropriate BASS add-on not loaded), 0 = not packed.
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: BASS_MIDI_FontGetPreset, BASS_MIDI_FONTINFO structure
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontGetInfo($handle)
	Local $tStruct = DllStructCreate('PTR;PTR;PTR;DWORD;DWORD;DWORD;DWORD;')
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontGetInfo", "DWORD", $handle, "PTR", DllStructGetPtr($tStruct))
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Local $aReturn[7]
	For $i = 0 To 2
		$aReturn[$i] = _BASS_PtrStringRead(DllStructGetData($tStruct, $i + 1))
	Next
	For $i = 3 To 6
		$aReturn[$i] = DllStructGetData($tStruct, $i + 1)
	Next
	Return $aReturn
EndFunc   ;==>_BASS_MIDI_FontGetInfo



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontGetPreset
; Description ...: Retrieves the name of a preset in a soundfont.
; Syntax ........: _BASS_MIDI_FontGetPreset($handle, $preset, $bank)
; Parameters ....: - $handle - The soundfont to get the preset name from.
;                  - $preset - Preset number... -1 = any preset (the first encountered).
;                  - $bank - Bank number... -1 = any bank (the first encountered).
; Return values .: Success - If successful, the requested preset name is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Drum sets are located in bank 128.
; Related .......: BASS_MIDI_FontGetInfo
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontGetPreset($handle, $preset, $bank)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "PTR", "BASS_MIDI_FontGetPreset", "DWORD", $handle, "LONG", $preset, "LONG", $bank)
	If @error Then Return SetError(1, 1, 0)
	If Not $BASS_ret_[0] Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return _BASS_PtrStringRead($BASS_ret_[0])
EndFunc   ;==>_BASS_MIDI_FontGetPreset



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontGetVolume
; Description ...: Retrieves a soundfont's volume level.
; Syntax ........: _BASS_MIDI_FontGetVolume($handle)
; Parameters ....: - $handle - The soundfont to get the volume of.
; Return values .: Success - If successful, the soundfont's volume level is returned
;                  Failure - Returns -1 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: BASS_MIDI_FontSetVolume
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontGetVolume($handle)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "SINGLE", "BASS_MIDI_FontGetVolume", "DWORD", $handle)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] < 0 Then Return SetError(_BASS_ErrorGetCode(), 0, -1)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontGetVolume



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontInit
; Description ...: Initializes a soundfont.
; Syntax ........: _BASS_MIDI_FontInit($fname, $flags = 0)
; Parameters ....: - $file - Filename.
;                  - $flags - Any combination of these flags.
;                             BASS_UNICODE file is in UTF-16 form. Otherwise it is ANSI on Windows and UTF-8 on OSX.
; Return values .: Success - If successful, the soundfont's handle is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: BASSMIDI uses SF2 soundfonts to provide the sounds to use in the rendering of MIDI files. Several SF2 soundfonts can be found on the internet, including one (Chorium) on the BASS website.
;                  A soundfont needs to be initialized before it can be used to render MIDI streams. Once initialized, a soundfont can be assigned to MIDI streams using the BASS_MIDI_StreamSetFonts function. A single soundfont can be shared by multiple MIDI streams.
;                  Information on the initialized soundfont can be retrieved using BASS_MIDI_FontGetInfo.
;                  If the specified soundfont has already been initialized, the existing handle will be returned.
; Related .......: BASS_MIDI_FontFree, BASS_MIDI_FontGetInfo, BASS_MIDI_FontGetPreset, BASS_MIDI_FontPack, BASS_MIDI_FontLoad, BASS_MIDI_FontSetVolume, BASS_MIDI_StreamSetFonts, BASS_CONFIG_MIDI_COMPACT
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontInit($fname, $flags = 0)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_FontInit", "WSTR", $fname, "DWORD", BitOR($flags, $BASS_UNICODE))
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontInit



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontLoad
; Description ...: Preloads presets from a soundfont.
; Syntax ........: _BASS_MIDI_FontLoad($handle, $preset = -1, $bank = -1)
; Parameters ....: - $handle - The soundfont handle... 0 = all soundfonts.
;                  - $preset - Preset number to load... -1 = all presets.
;                  - $bank - Bank number to load... -1 = all banks.
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Samples are normally loaded as they are needed while rendering a MIDI stream, which can result in CPU spikes, particularly with packed soundfonts. That generally won't cause any problems, but when smooth/constant performance is critical this function can be used to preload the samples before rendering, so avoiding the need to load them while rendering.
;                  When preloading samples to render a particular MIDI stream, it is more efficient to use BASS_MIDI_StreamLoadSamples to preload the specific samples that the MIDI stream will use, rather than preloading the entire soundfont.
;                  Preloaded samples can be compacted/unloaded just like any other samples, so it is probably wise to disable the BASS_CONFIG_MIDI_COMPACT option when preloading samples, to avoid any chance of the samples subsequently being automatically unloaded.
;                  A soundfont should not be preloaded while it's being used to render any MIDI streams, as that could delay the rendering.
; Related .......: BASS_MIDI_FontCompact, BASS_MIDI_FontGetInfo, BASS_MIDI_FontGetPreset, BASS_MIDI_FontInit, BASS_MIDI_StreamLoadSamples, BASS_CONFIG_MIDI_COMPACT
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontLoad($handle, $preset = -1, $bank = -1)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontLoad", "DWORD", $handle, "LONG", $preset, "LONG", $bank)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontLoad



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontPack
; Description ...: Produces a compressed version of a soundfont.
; Syntax ........: _BASS_MIDI_FontPack($handle, $outfile, $encoder, $flags)
; Parameters ....: - $handle - The soundfont to pack.
;                  - $outfile - Filename for the packed soundfont.
;                  - $encoder - Encoder command-line.
;                  - $flags - Any combination of these flags.
;                             BASS_UNICODE outfile and encoder are in UTF-16 form. Otherwise they are ANSI on Windows and UTF-8 on OSX.
; Return values .: Success - If successful, the TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Standard soundfonts use PCM samples, so they can be quite large, which can be a problem if they're to be distributed. To reduce the size, BASSMIDI can compress the samples using any command-line encoder with STDIN & STDOUT support. Packed soundfonts can be used for rendering by BASSMIDI just like normal soundfonts. They can also be unpacked using BASS_MIDI_FontUnpack.
;                  Although any command-line encoder can be used, it is best to use a lossless format like FLAC or WavPack, rather than a lossy one like OGG or MP3. Using a lossless encoder, the packed soundfont will produce exactly the same results as the original soundfont, and will be identical to the original when unpacked. As a compromise between quality and size, the WavPack hybrid/lossy mode also produces good sounding results.
;                  The encoder must be told (via the command-line) to expect input from STDIN and to send it's output to STDOUT.
;                  To use a packed soundfont, the appropriate BASS add-on needs to be loaded (via BASS_PluginLoad) to decode the samples. For example, if the samples are FLAC encoded, BASSFLAC would need to be loaded. During rendering, the samples are unpacked as they're needed, which could result in CPU spikes. Where smooth performance is critical, it may be wise to preload the samples using BASS_MIDI_FontLoad or BASS_MIDI_StreamLoadSamples.
;                  A soundfont should not be packed while it's being used to render any MIDI streams, as that could delay the rendering.
; Related .......: BASS_MIDI_FontInit, BASS_MIDI_FontLoad, BASS_MIDI_FontUnpack
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontPack($handle, $outfile, $encoder, $flags)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontPack", "DWORD", $handle, "WSTR", $outfile, "WSTR", $encoder, "DWORD", BitOR($flags, $BASS_UNICODE))
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontPack



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontSetVolume
; Description ...: Sets a soundfont's volume level.
; Syntax ........: _BASS_MIDI_FontSetVolume($handle, $volume)
; Parameters ....: - $handle - The soundfont to set the volume of.
;                  - $volume - The volume level... 0 = silent, 1.0 = normal/default.
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: By default, some soundfonts may be louder than others, which could be a problem when mixing multiple soundfonts. This function can be used to compensate for any differences, by raising the level of the quieter soundfonts or lowering the louder ones.
;                  Changes take immediate effect in any MIDI streams that are using the soundfont.
; Related .......: BASS_MIDI_FontGetVolume
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontSetVolume($handle, $volume)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontSetVolume", "DWORD", $handle, "SINGLE", $volume)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontSetVolume



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_FontUnpack
; Description ...: Produces a decompressed version of a packed soundfont.
; Syntax ........: _BASS_MIDI_FontUnpack($handle, $outfile, $flags)
; Parameters ....: - $handle - The soundfont to unpack.
;                  - $outfile - Filename for the unpacked soundfont.
;                  - $flags - Any combination of these flags.
;                             BASS_UNICODE outfile is in UTF-16 form. Otherwise it is ANSI on Windows and UTF-8 on OSX.
; Return values .: Success - If successful, the TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: To unpack a soundfont, the appropriate BASS add-on needs to be loaded (via BASS_PluginLoad) to decode the samples. For example, if the samples are FLAC encoded, BASSFLAC would need to be loaded. BASS also needs to have been initialized, using BASS_Init. For just unpacking a soundfont, the "no sound" device could be used.
;                  A soundfont should not be unpacked while it's being used to render any MIDI streams, as that could delay the rendering.
;                  BASS_MIDI_FontGetInfo can be used to check if a soundfont is packed.
; Related .......: BASS_MIDI_FontInit, BASS_MIDI_FontPack
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_FontUnpack($handle, $outfile, $flags)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_FontUnpack", "DWORD", $handle, "WSTR", $outfile, "DWORD", BitOR($flags, $BASS_UNICODE))
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_FontUnpack



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_InFree
; Description ...: Frees a MIDI input device.
; Syntax ........: _BASS_MIDI_InFree($device)
; Parameters ....: - $device - The device to free.
; Return values .: Success - If successful, the TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: _BASS_MIDI_InInit
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_InFree($device)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_InFree", "DWORD", $device)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_InFree



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_InGetDeviceInfo
; Description ...: Retrieves information on an output device.
; Syntax ........: _BASS_MIDI_InGetDeviceInfo($device)
; Parameters ....: -	$device		-	The device to get the information of... 0 = first.
; Return values .: Success      - Returns Array of Device information.
;									- [0] = Name - Description of the device.
;									- [1] = id - An identification number.
;									- [2] =  The device's current status... a combination of these flags. BASS_DEVICE_ENABLED The device is enabled. It will not be possible to initialize the device if this flag is not present.
;                                            BASS_DEVICE_INIT The device is initialized, ie. BASS_MIDI_InInit has been called.
;                  Failure      - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
;									@error will be set to-
;										- $BASS_ERROR_DEVICE device is invalid.
; Author ........: Brett Francis (BrettF)
; Modified ......: Eukalyptus
; Remarks .......:
; Related .......: _BASS_MIDI_InInit
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_InGetDeviceInfo($device)
	Local $aRet[3]
	Local $BASS_ret_struct = DllStructCreate("ptr name;dword id;dword flags;")
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "bool", "BASS_MIDI_InGetGeviceInfo", "dword", $device, "ptr", DllStructGetPtr($BASS_ret_struct))
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)

	$aRet[0] = _BASS_PtrStringRead(DllStructGetData($BASS_ret_struct, 1))
	$aRet[1] = DllStructGetData($BASS_ret_struct, 2)
	$aRet[2] = DllStructGetData($BASS_ret_struct, 3)
	Return $aRet
EndFunc   ;==>_BASS_MIDI_InGetGeviceInfo



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_InInit
; Description ...: Initializes a MIDI input device.
; Syntax ........: _BASS_MIDI_InInit($device, $proc = 0, $user = 0)
; Parameters ....: - $device - The device to use... 0 = first. BASS_MIDI_InGetDeviceInfo can be used to enumerate the available devices.
;                  - $proc - Callback function to receive MIDI data from the device.
;                  - $user - User instance data to pass to the callback function.
; Return values .: Success - If the device was successfully initialized, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: _BASS_MIDI_InFree, _BASS_MIDI_InGetDeviceInfo, _BASS_MIDI_InStart
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_InInit($device, $proc = 0, $user = 0)
	Local $proc_s = -1
	If IsString($proc) Then
		$proc_s = DllCallbackRegister($proc, "ptr", "dword;double;ptr;dword;ptr")
		$proc = DllCallbackGetPtr($proc_s)
	EndIf
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_InInit", "DWORD", $device, "PTR", $proc, "PTR", $user)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then
		If $proc_s <> -1 Then DllCallbackFree($proc_s)
		Return SetError(_BASS_ErrorGetCode(), 0, 0)
	EndIf
	Return SetExtended($proc_s, $BASS_ret_[0])
EndFunc   ;==>_BASS_MIDI_InInit



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_InStart
; Description ...: Starts a MIDI input device.
; Syntax ........: _BASS_MIDI_InStart($device)
; Parameters ....: - $device - The device to free.
; Return values .: Success - If successful, the TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: _BASS_MIDI_InStop
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_InStart($device)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_InStart", "DWORD", $device)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_InStart



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_InStop
; Description ...: Stops a MIDI input device.
; Syntax ........: _BASS_MIDI_InStop($device)
; Parameters ....: - $device - The device to free.
; Return values .: Success - If successful, the TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: _BASS_MIDI_InFree, _BASS_MIDI_InStart
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_InStop($device)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_InStop", "DWORD", $device)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_InStop



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamCreate
; Description ...: Creates a sample stream to render real-time MIDI events.
; Syntax ........: _BASS_MIDI_StreamCreate($channels, $flags = 0, $freq = 0)
; Parameters ....: - $channels - The number of MIDI channels... 1 (min) - 128 (max).
;                  - $flags - Any combination of these flags.
;                             BASS_SAMPLE_8BITS Use 8-bit resolution. If neither this or the BASS_SAMPLE_FLOAT flags are specified, then the sample data will be 16-bit.
;                             BASS_SAMPLE_FLOAT Use 32-bit floating-point sample data. See Floating-point channels for more info.
;                             BASS_SAMPLE_MONO Decode/play the MIDI in mono (uses less CPU than stereo). This flag is automatically applied if BASS_DEVICE_MONO was specified when calling BASS_Init.
;                             BASS_SAMPLE_SOFTWARE Force the stream to not use hardware mixing.
;                             BASS_SAMPLE_3D Enable 3D functionality. This requires that the BASS_DEVICE_3D flag was specified when calling BASS_Init. 3D channels must also be mono, so BASS_SAMPLE_MONO is automatically applied. The SPEAKER flags cannot be used together with this flag.
;                             BASS_SAMPLE_FX requires DirectX 8 or aboveEnable the old implementation of DirectX 8 effects. See the DX8 effect implementations section for details. Use BASS_ChannelSetFX to add effects to the stream.
;                             BASS_STREAM_AUTOFREE Automatically free the stream when playback ends.
;                             BASS_STREAM_DECODE Decode/render the sample data, without playing it. Use BASS_ChannelGetData to retrieve decoded sample data. The BASS_SAMPLE_3D, BASS_STREAM_AUTOFREE and SPEAKER flags cannot be used together with this flag. The BASS_SAMPLE_SOFTWARE and BASS_SAMPLE_FX flags are also ignored.
;                             BASS_SPEAKER_xxx Speaker assignment flags. The BASS_SAMPLE_MONO flag is automatically applied when using a mono speaker assignment flag.
;                             BASS_MIDI_NOFX Disable reverb and chorus processing, saving some CPU time. This flag can be toggled at any time using BASS_ChannelFlags.
; Return values .: Success - If successful, the new stream's handle is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: This function creates a stream solely for real-time MIDI events. As it's not based on any file, the stream has no predetermined length and is never-ending. Seeking isn't possible, but it is possible to reset everything, including playback buffer, by calling BASS_ChannelPlay (restart = TRUE) or BASS_ChannelSetPosition (pos = 0).
;                  MIDI events are applied using the BASS_MIDI_StreamEvent function. If the stream is being played (it's not a decoding channel), then there will be some delay in the effect of the events being heard. This latency can be reduced by making use of the BASS_CONFIG_BUFFER and BASS_CONFIG_UPDATEPERIOD config options.
;                  Unlike MIDI files, streams created for rendering real-time events are not limited to 16 channels. By default, all the MIDI channels are melodic, but they can be switched to percussion/drums using BASS_MIDI_StreamEvent and the MIDI_EVENT_DRUMS event.
;                  Soundfonts provide the sounds that are used to render a MIDI stream. A default soundfont configuration is applied initially to the new MIDI stream, which can subsequently be overridden using BASS_MIDI_StreamSetFonts.
;                  To play a MIDI file, use BASS_MIDI_StreamCreateFile.
; Related .......: BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamEvent, BASS_MIDI_StreamGetChannel, BASS_MIDI_StreamSetFonts, BASS_CONFIG_MIDI_DEFFONT, BASS_CONFIG_MIDI_VOICES, BASS_ChannelGetInfo, BASS_ChannelGetLength, BASS_ChannelGetTags, BASS_ChannelPlay, BASS_ChannelSetAttribute, BASS_ChannelSetDSP, BASS_ChannelSetFX, BASS_ChannelSetLink, BASS_StreamFree
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamCreate($channels, $flags = 0, $freq = 0)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamCreate", "DWORD", $channels, "DWORD", $flags, "DWORD", $freq)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamCreate



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamCreateFile
; Description ...: Creates a sample stream from a MIDI file.
; Syntax ........: _BASS_MIDI_StreamCreateFile($mem, $file, $offset = 0, $length = 0, $flags = 0, $freq = 0)
; Parameters ....: - $mem - TRUE = stream the file from memory.
;                  - $file - Filename (mem = FALSE) or a memory location (mem = TRUE).
;                  - $offset - File offset to begin streaming from (only used if mem = FALSE).
;                  - $length - Data length... 0 = use all data up to the end of the file (if mem = FALSE).
;                  - $flags - Any combination of these flags.
;                             BASS_SAMPLE_8BITS Use 8-bit resolution. If neither this or the BASS_SAMPLE_FLOAT flags are specified, then the sample data will be 16-bit.
;                             BASS_SAMPLE_FLOAT Use 32-bit floating-point sample data. See Floating-point channels for more info.
;                             BASS_SAMPLE_MONO Decode/play the MIDI in mono (uses less CPU than stereo). This flag is automatically applied if BASS_DEVICE_MONO was specified when calling BASS_Init.
;                             BASS_SAMPLE_SOFTWARE Force the stream to not use hardware mixing.
;                             BASS_SAMPLE_3D Enable 3D functionality. This requires that the BASS_DEVICE_3D flag was specified when calling BASS_Init. 3D channels must also be mono, so BASS_SAMPLE_MONO is automatically applied. The SPEAKER flags cannot be used together with this flag.
;                             BASS_SAMPLE_LOOP Loop the file. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_SAMPLE_FX requires DirectX 8 or aboveEnable the old implementation of DirectX 8 effects. See the DX8 effect implementations section for details. Use BASS_ChannelSetFX to add effects to the stream.
;                             BASS_STREAM_AUTOFREE Automatically free the stream when playback ends.
;                             BASS_STREAM_DECODE Decode/render the sample data, without playing it. Use BASS_ChannelGetData to retrieve decoded sample data. The BASS_SAMPLE_3D, BASS_STREAM_AUTOFREE and SPEAKER flags cannot be used together with this flag. The BASS_SAMPLE_SOFTWARE and BASS_SAMPLE_FX flags are also ignored.
;                             BASS_SPEAKER_xxx Speaker assignment flags. The BASS_SAMPLE_MONO flag is automatically applied when using a mono speaker assignment flag.
;                             BASS_MIDI_DECAYEND Let the sound decay naturally (including reverb) instead of stopping it abruptly at the end of the file. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_MIDI_DECAYSEEK Let the old sound decay naturally (including reverb) when changing the position, including looping. This flag can be toggled at any time using BASS_ChannelFlags, and can also be used in BASS_ChannelSetPosition calls to have it apply to particular position changes.
;                             BASS_MIDI_NOFX Disable reverb and chorus processing, saving some CPU time. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_UNICODE file is in UTF-16 form. Otherwise it is ANSI on Windows and UTF-8 on OSX.
; Return values .: Success - If successful, the new stream's handle is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: BASSMIDI supports standard MIDI format 0/1/2 files. In the case of format 2, the tracks are rendered/played one after another. RIFF MIDI (RMID) files are also supported. The General MIDI standard events are supported, as are several Roland GS and Yamaha XG NRPN and SysEx events. A full list of supported MIDI events can be found in the BASS_MIDI_StreamEvent documentation.
;                  Soundfonts provide the sounds that are used to render a MIDI stream. A default soundfont configuration is applied initially to the new MIDI stream, which can subsequently be overridden using BASS_MIDI_StreamSetFonts. By default, with the BASS_CONFIG_MIDI_AUTOFONT config option enabled, BASSMIDI will also check for a soundfont of the same name as the MIDI file. Note that a MIDI stream can have multiple soundfonts stacked, each providing different presets, for example.
;                  As well as the standard byte/time-based positioning, MIDI tick-based positioning is also supported. The BASS_POS_MIDI_TICK "mode" can be used with BASS_ChannelGetLength, BASS_ChannelGetPosition and BASS_ChannelSetPosition to deal in ticks. The BASS_ATTRIB_MIDI_PPQN attribute can be used to translate the position to beats.
;                  Marker, cue, text, and lyric events can be retrieved via the BASS_MIDI_StreamGetMark function. Syncs can also be used to be notified of their occurrence.
;                  Other texts of each track (eg. track name) are available via the BASS_TAG_MIDI_TRACK+<track> tag, where track=0 is the first track. A pointer to a series of null-terminated strings is given, the final string ending with a double null. The first text in the first track is generally the title of the MIDI file. RIFF MIDI tags are also available via the standard BASS_TAG_RIFF_INFO tag.
;                  Unlike with most stream formats, the entire MIDI file is loaded to memory. This means the file can be deleted or moved after calling this function, or the memory can be discarded (mem = TRUE).
;                  To play a MIDI file from the internet, use BASS_MIDI_StreamCreateURL.
; Related .......: BASS_MIDI_StreamCreate, BASS_MIDI_StreamCreateFileUser, BASS_MIDI_StreamCreateURL, BASS_MIDI_StreamGetChannel, BASS_MIDI_StreamGetEvents, BASS_MIDI_StreamGetMark, BASS_MIDI_StreamLoadSamples, BASS_MIDI_StreamSetFonts, BASS_CONFIG_MIDI_AUTOFONT, BASS_CONFIG_MIDI_DEFFONT, BASS_CONFIG_MIDI_VOICES
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamCreateFile($mem, $file, $offset = 0, $length = 0, $flags = 0, $freq = 0)
	Local $tpFile = "ptr"
	If IsString($file) Then $tpFile = "wstr"
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamCreateFile", "int", $mem, $tpFile, $file, "uint64", $offset, "uint64", $length, "DWORD", BitOR($flags, $BASS_UNICODE), "DWORD", $freq)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamCreateFile



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamCreateFileUser
; Description ...: Creates a sample stream from a MIDI file via user callback functions.
; Syntax ........: _BASS_MIDI_StreamCreateFileUser($system, $flags = 0, $procs = 0, $user = 0, $freq = 0)
; Parameters ....: - $system - File system to use, which must be STREAMFILE_NOBUFFER, as the entire MIDI file is preloaded.
;                  - $flags - Any combination of these flags.
;                             BASS_SAMPLE_8BITS Use 8-bit resolution. If neither this or the BASS_SAMPLE_FLOAT flags are specified, then the sample data will be 16-bit.
;                             BASS_SAMPLE_FLOAT Use 32-bit floating-point sample data. See Floating-point channels for more info.
;                             BASS_SAMPLE_MONO Decode/play the MIDI in mono (uses less CPU than stereo). This flag is automatically applied if BASS_DEVICE_MONO was specified when calling BASS_Init.
;                             BASS_SAMPLE_SOFTWARE Force the stream to not use hardware mixing.
;                             BASS_SAMPLE_3D Enable 3D functionality. This requires that the BASS_DEVICE_3D flag was specified when calling BASS_Init. 3D channels must also be mono, so BASS_SAMPLE_MONO is automatically applied. The SPEAKER flags cannot be used together with this flag.
;                             BASS_SAMPLE_LOOP Loop the file. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_SAMPLE_FX requires DirectX 8 or aboveEnable the old implementation of DirectX 8 effects. See the DX8 effect implementations section for details. Use BASS_ChannelSetFX to add effects to the stream.
;                             BASS_STREAM_AUTOFREE Automatically free the stream when playback ends.
;                             BASS_STREAM_DECODE Decode/render the sample data, without playing it. Use BASS_ChannelGetData to retrieve decoded sample data. The BASS_SAMPLE_3D, BASS_STREAM_AUTOFREE and SPEAKER flags cannot be used together with this flag. The BASS_SAMPLE_SOFTWARE and BASS_SAMPLE_FX flags are also ignored.
;                             BASS_SPEAKER_xxx Speaker assignment flags. The BASS_SAMPLE_MONO flag is automatically applied when using a mono speaker assignment flag.
;                             BASS_MIDI_DECAYEND Let the sound decay naturally (including reverb) instead of stopping it abruptly at the end of the file. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_MIDI_DECAYSEEK Let the old sound decay naturally (including reverb) when changing the position, including looping. This flag can be toggled at any time using BASS_ChannelFlags, and can also be used in BASS_ChannelSetPosition calls to have it apply to particular position changes.
;                             BASS_MIDI_NOFX Disable reverb and chorus processing, saving some CPU time. This flag can be toggled at any time using BASS_ChannelFlags.
; Return values .: Success - If successful, the new stream's handle is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: As there is no file associated with a user file stream, it's not possible for BASSMIDI to look for a soundfont with the same name as the MIDI file. If there is a matching soundfont, it can be applied using BASS_MIDI_StreamSetFonts.
; Related .......: BASS_MIDI_StreamCreate, BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamCreateURL, BASS_MIDI_StreamGetChannel, BASS_MIDI_StreamGetMark, BASS_MIDI_StreamLoadSamples, BASS_MIDI_StreamSetFonts, BASS_CONFIG_MIDI_DEFFONT, BASS_CONFIG_MIDI_VOICES, BASS_ChannelGetInfo, BASS_ChannelGetLength, BASS_ChannelGetTags, BASS_ChannelPlay, BASS_ChannelSetAttribute, BASS_ChannelSetDSP, BASS_ChannelSetFX, BASS_ChannelSetLink, BASS_StreamFree, BASS_FILEPROCS structure
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamCreateFileUser($system, $flags = 0, $procs = 0, $user = 0, $freq = 0)
	Local $proc_s = -1
	If IsString($procs) Then
		$proc_s = DllCallbackRegister($procs, "ptr", "ptr;ptr;ptr;ptr")
		$procs = DllCallbackGetPtr($proc_s)
	EndIf
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamCreateFileUser", "DWORD", $system, "DWORD", $flags, "PTR", $procs, "PTR", $user, "DWORD", $freq)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then
		If $proc_s <> -1 Then DllCallbackFree($proc_s)
		Return SetError(_BASS_ErrorGetCode(), 0, 0)
	EndIf
	Return SetExtended($proc_s, $BASS_ret_[0])
EndFunc   ;==>_BASS_MIDI_StreamCreateFileUser



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamCreateURL
; Description ...: Creates a sample stream from an MIDI file on the internet, optionally receiving the downloaded data in a callback.
; Syntax ........: _BASS_MIDI_StreamCreateURL($URL, $offset = 0, $flags = 0, $proc = 0, $user = 0, $freq = 0)
; Parameters ....: - $url - URL of the file to stream. Should begin with "http://" or "ftp://".
;                  - $offset - File position to start streaming from. This is ignored by some servers, specifically when the file length is unknown.
;                  - $flags - Any combination of these flags.
;                             BASS_SAMPLE_8BITS Use 8-bit resolution. If neither this or the BASS_SAMPLE_FLOAT flags are specified, then the sample data will be 16-bit.
;                             BASS_SAMPLE_FLOAT Use 32-bit floating-point sample data. See Floating-point channels for more info.
;                             BASS_SAMPLE_MONO Decode/play the MIDI in mono (uses less CPU than stereo). This flag is automatically applied if BASS_DEVICE_MONO was specified when calling BASS_Init.
;                             BASS_SAMPLE_SOFTWARE Force the stream to not use hardware mixing.
;                             BASS_SAMPLE_3D Enable 3D functionality. This requires that the BASS_DEVICE_3D flag was specified when calling BASS_Init. 3D channels must also be mono, so BASS_SAMPLE_MONO is automatically applied. The SPEAKER flags cannot be used together with this flag.
;                             BASS_SAMPLE_LOOP Loop the file. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_SAMPLE_FX requires DirectX 8 or aboveEnable the old implementation of DirectX 8 effects. See the DX8 effect implementations section for details. Use BASS_ChannelSetFX to add effects to the stream.
;                             BASS_STREAM_STATUS Pass status info (HTTP/ICY tags) from the server to the DOWNLOADPROC callback during connection. This can be useful to determine the reason for a failure.
;                             BASS_STREAM_AUTOFREE Automatically free the stream when playback ends.
;                             BASS_STREAM_DECODE Decode/render the sample data, without playing it. Use BASS_ChannelGetData to retrieve decoded sample data. The BASS_SAMPLE_3D, BASS_STREAM_AUTOFREE and SPEAKER flags cannot be used together with this flag. The BASS_SAMPLE_SOFTWARE and BASS_SAMPLE_FX flags are also ignored.
;                             BASS_SPEAKER_xxx Speaker assignment flags. The BASS_SAMPLE_MONO flag is automatically applied when using a mono speaker assignment flag.
;                             BASS_MIDI_DECAYEND Let the sound decay naturally (including reverb) instead of stopping it abruptly at the end of the file. This flag can be toggled at any time using BASS_ChannelFlags.
;                             BASS_MIDI_DECAYSEEK Let the old sound decay naturally (including reverb) when changing the position, including looping. This flag can be toggled at any time using BASS_ChannelFlags, and can also be used in BASS_ChannelSetPosition calls to have it apply to particular position changes.
;                             BASS_MIDI_NOFX Disable reverb and chorus processing, saving some CPU time. This flag can be toggled at any time using BASS_ChannelFlags.
; Return values .: Success - If successful, the new stream's handle is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: The entire MIDI file is preloaded, so the standard BASS_STREAM_BLOCK and BASS_STREAM_RESTRATE flags have no effect here.
;                  Regardless of the BASS_CONFIG_MIDI_AUTOFONT setting, a matching soundfont is not looked for when opening a MIDI file from a URL.
; Related .......: BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamCreateFileUser, BASS_MIDI_StreamGetChannel, BASS_MIDI_StreamGetMark, BASS_MIDI_StreamLoadSamples, BASS_MIDI_StreamSetFonts, BASS_CONFIG_MIDI_VOICES, BASS_ChannelGetInfo, BASS_ChannelGetLength, BASS_ChannelGetTags, BASS_ChannelPlay, BASS_ChannelSetAttribute, BASS_ChannelSetDSP, BASS_ChannelSetFX, BASS_ChannelSetLink, BASS_StreamFree, DOWNLOADPROC callback, BASS_CONFIG_NET_AGENT, BASS_CONFIG_NET_PROXY, BASS_CONFIG_NET_TIMEOUT
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamCreateURL($URL, $offset = 0, $flags = 0, $proc = 0, $user = 0, $freq = 0)
	Local $proc_s = -1
	If IsString($proc) Then
		$proc_s = DllCallbackRegister($proc, "ptr", "ptr;ptr;ptr;ptr")
		$proc = DllCallbackGetPtr($proc_s)
	EndIf
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamCreateURL", "STR", $URL, "DWORD", $offset, "DWORD", $flags, "PTR", $proc, "PTR", $user, "DWORD", $freq)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then
		If $proc_s <> -1 Then DllCallbackFree($proc_s)
		Return SetError(_BASS_ErrorGetCode(), 0, 0)
	EndIf
	Return SetExtended($proc_s, $BASS_ret_[0])
EndFunc   ;==>_BASS_MIDI_StreamCreateURL



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamEvent
; Description ...: Applies an event to a channel in a MIDI stream.
; Syntax ........: _BASS_MIDI_StreamEvent($handle, $chan, $event, $param)
; Parameters ....: - $handle - The MIDI stream to apply the event to.
;                  - $chan - The MIDI channel to apply the event to... 0 = channel 1.
;                  - $event - The event to apply, see the table below.
;                  - $param - The event parameter.
; -------------------------------------------------------------------------------------------------------------------------------
; 						Event types, with param definitions.
; 							$MIDI_EVENT_NOTE			- Press or release a key.
; 															- param : LOBYTE = key number (0-127, 60=middle C), HIBYTE = velocity (0=release, 1-127=press).
; 							$MIDI_EVENT_PROGRAM			- Select the preset/instrument to use. Standard soundfont presets follow the General MIDI standard, and generally also include Roland GS variations in other banks (accessible via the MIDI_EVENT_BANK event).
; 															- param : preset number (0-127).
; 							$MIDI_EVENT_CHANPRES		- Set the channel pressure.
; 															- param : pressure level (0-127).
; 							$MIDI_EVENT_PITCH			- Set the pitch wheel position.
; 															- param : pitch wheel position (0-16383, 8192=normal/middle).
; 							$MIDI_EVENT_BANK			- Select the bank to use (MIDI controller 0).
; 															- param : bank number (0-127).
; 							$MIDI_EVENT_MODULATION		- Set the modulation (MIDI controller 1)
; 															- param : modulation level (0-127).
; 							$MIDI_EVENT_PORTATIME		- Set the portamento time (MIDI controller 5).
; 															- param : portamento time (0-127).
; 							$MIDI_EVENT_VOLUME			- Set the volume (MIDI controller 7).
; 															- param : volume level (0-127).
; 							$MIDI_EVENT_PAN				- Set the pan position (MIDI controller 10).
; 															- param : pan position (0-127, 64=middle).
; 							$MIDI_EVENT_EXPRESSION		- Set the expression (MIDI controller 11).
; 															- param : expression level (0-127).
; 							$MIDI_EVENT_SUSTAIN			- Set the sustain switch (MIDI controller 64).
; 															- param : enable sustain? (0-63=no, 64-127=yes).
; 							$MIDI_EVENT_PORTAMENTO		- Set the portamento switch (MIDI controller 65).
; 															- param : enable portamento? (0-63=no, 64-127=yes).
; 							$MIDI_EVENT_RESONANCE		- Set the low-pass filter resonance (MIDI controller 71, NRPN 121h)
; 															- param : resonance level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_RELEASE			- Set the release time (MIDI controller 72, NRPN 166h)
; 															- param : release time (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_ATTACK			- Set the attack time (MIDI controller 73, NRPN 163h)
; 															- param : attack time (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_CUTOFF			- Set the low-pass filter cutoff (MIDI controller 74, NRPN 120h)
; 															- param : cutoff level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_PORTANOTE		- Set the portamento start key; the next note starts at this key (MIDI controller 84).
; 															- param : key number (1-127, 60=middle C).
; 							$MIDI_EVENT_REVERB			- Set the reverb send level (MIDI controller 91)
; 															- param : reverb level (0-127).
; 							$MIDI_EVENT_CHORUS			- Set the chorus send level (MIDI controller 93)
; 															- param : chorus level (0-127).
; 							$MIDI_EVENT_SOUNDOFF		- Stop all sounds (MIDI controller 120).
; 															- param : not used.
; 							$MIDI_EVENT_RESET			- Reset controllers (MIDI controller 121), that is modulation=0, expression=127, sustain=0, portamento=0, release time=64, attack time=64, pitch wheel=8192, channel pressure=0.
; 															- param : not used.
; 							$MIDI_EVENT_NOTESOFF		- Release all keys (MIDI controller 123).
; 															- param : not used.
; 							$MIDI_EVENT_MODE			- Set poly/mono mode (MIDI controllers 126 & 127).
; 															- param : mode (0=poly, 1=mono).
; 							$MIDI_EVENT_PITCHRANGE		- Set pitch wheel range (MIDI RPN 0).
; 															- param : range in semitones.
; 							$MIDI_EVENT_FINETUNE		- Set the fine tuning (MIDI RPN 1).
; 															- param : finetune in cents (0-16383, 0=-100, 8192=normal, 16383=+100).
; 							$MIDI_EVENT_COARSETUNE		- Set the coarse tuning (MIDI RPN 2).
; 															- param : finetune in semitones (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUMS			- Set the percussion/drums channel switch.
; 															- param : use drums? (0=no, 1=yes).
; 							$MIDI_EVENT_DRUM_CUTOFF		- Set the low-pass filter cutoff of a drum key (MIDI NRPN 14knh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = cutoff level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUM_RESONANCE	- Set the low-pass filter resonance of a drum key (MIDI NRPN 15knh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = resonance level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUM_COARSETUNE	- Set the coarse tuning of a drum key (MIDI NRPN 18knh).
; 															- param : LOBYTE = key number (0-127), HIBYTE = finetune in semitones (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUM_FINETUNE	- Set the fine tuning of a drum key (MIDI NRPN 19knh).
; 															- param : LOBYTE = key number (0-127), HIBYTE = finetune in cents (0-127, 0=-100, 64=normal, 127=+100).
; 							$MIDI_EVENT_DRUM_LEVEL		- Set the level of a drum key (MIDI NRPN 1Aknh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = level (0-127, 127=normal/full).
; 							$MIDI_EVENT_DRUM_PAN		- Set the pan position of a drum key (MIDI NRPN 1Cknh).
; 															- param : LOBYTE = key number (0-127), HIBYTE = pan position (0-127, 0=random, 64=middle).
; 							$MIDI_EVENT_DRUM_REVERB		- Set the reverb send level of a drum key (MIDI NRPN 1Dknh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = reverb level (0-127).
; 							$MIDI_EVENT_DRUM_CHORUS		- Set the chorus send level of a drum key (MIDI NRPN 1Eknh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = chorus level (0-127).
; 						Global events.
; 							$MIDI_EVENT_TEMPO			- Set the tempo (MIDI meta event 81). Changing the tempo affects the stream length, and the BASS_ChannelGetLength byte value will no longer be valid.
; 															- param : tempo in microseconds per quarter note.
; 							$MIDI_EVENT_MASTERVOL		- Set the master volume.
; 															- param : volume level (0-16383, 0=silent, 16363=normal/full).
; 							$MIDI_EVENT_REVERB_TIME		- Set the reverb time.
; 															- param : reverb time in milliseconds.
; 							$MIDI_EVENT_REVERB_DELAY	- Set the reverb delay.
; 															- param : reverb delay in millisecond 10ths.
; 							$MIDI_EVENT_REVERB_LOCUTOFF	- Set the reverb low-pass cutoff
; 															- param : reverb low-pass cutoff in hertz (0=off).
; 							$MIDI_EVENT_REVERB_HICUTOFF	- Set the reverb high-pass cutoff
; 															- param : reverb high-pass cutoff in hertz (0=off).
; 							$MIDI_EVENT_REVERB_LEVEL	- Set the reverb level
; 															- param : reverb level (0=off, 100=0dB, 200=+6dB).
; 							$MIDI_EVENT_CHORUS_DELAY	- Set the chorus delay.
; 															- param : chorus delay in millisecond 10ths.
; 							$MIDI_EVENT_CHORUS_DEPTH	- Set the chorus depth.
; 															- param : chorus depth in millisecond 10ths.
; 							$MIDI_EVENT_CHORUS_RATE		- Set the chorus rate.
; 															- param : chorus rate in hertz 100ths.
; 							$MIDI_EVENT_CHORUS_FEEDBACK	- Set the chorus feedback level.
; 															- param : chorus feedback level (0=-100%, 100=off, 200=+100%).
; 							$MIDI_EVENT_CHORUS_LEVEL	- Set the chorus level.
; 															- param : chorus level (0=off, 100=0dB, 200=+6dB).
; 							$MIDI_EVENT_CHORUS_REVERB	- Set the chorus send to reverb level.
; 															- param : chorus send to reverb level (0=off, 100=0dB, 200=+6dB).
; 						Other (non-MIDI) events.
; 							$MIDI_EVENT_MIXLEVEL		- Set the level.
; 															- param : the level (0=silent, 100=0dB, 200=+6dB).
; 							$MIDI_EVENT_TRANSPOSE		- Transpose all notes. Changes take effect from the next note played, and affect melodic channels only (not drum channels).
; 															- param : transposition amount in semitones (0=-100, 100=normal, 200=+100).
; -------------------------------------------------------------------------------------------------------------------------------
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Apart from the "global" events, all events apply only to the specified MIDI channel.
;                  Events applied to a MIDI file stream can subsequently be overridden by events in the file itself, and will also be overridden when seeking or looping. This applies to all events apart from the "non-MIDI" events. Syncs can be used to be informed of when event values change.
;                  If the MIDI stream is being played (it's not a decoding channel), then there will be some delay in the effect of the event being heard. This latency can be reduced by making use of the BASS_CONFIG_BUFFER and BASS_CONFIG_UPDATEPERIOD config options when creating the stream.
; Related .......: BASS_MIDI_StreamCreate, BASS_MIDI_StreamGetEvent
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamEvent($handle, $chan, $event, $param)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_StreamEvent", "DWORD", $handle, "DWORD", $chan, "DWORD", $event, "DWORD", $param)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamEvent



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamEvents
; Description ...: Applies any number of events to a MIDI stream.
; Syntax ........: _BASS_MIDI_StreamEvents($handle, $mode, $events, $length = 0)
; Parameters ....: - $handle - The MIDI stream to apply the events to.
;                  - $mode - The type of event data to apply. One of the following, with optional flags. BASS_MIDI_EVENTS_RAW Raw MIDI event data, as would be sent to a MIDI device. Running status is supported. To overcome the 16 channel limit, the event data's channel information can optionally be overridden by adding the new channel number to this parameter, where +1 = the 1st channel.
;                  - $events - The event data: can be an array as returned by _BASS_MIDI_StreamGetEvents, a DllStruct or a pointer to a DllStruct
;                  - $length - The length of the event data. The number of bytes or BASS_MIDI_EVENT structures, depending on the type of event data.
; Return values .: Success - If successful, the number of events processed is returned
;                  Failure - Returns -1 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: By default, MIDI channels do not have streams assigned to them; a MIDI channel only gets a stream when this function is called, which it then keeps until the MIDI stream is freed. MIDI channel streams can also be freed before then via BASS_StreamFree. Each MIDI channel stream increases the CPU usage slightly, even if there are no DSP/FX set on them, so for optimal performance they should not be activated when unnecessary.
;                  The MIDI channel streams have a different path to the final mix than the BASSMIDI reverb/chorus processing, which means that the reverb/chorus will not be present in the data received by any DSP/FX set on the streams and nor will the reverb/chorus be applied to the DSP/FX output; the reverb/chorus processing will use the channel's original data.
;                  The MIDI channel streams can only be used to set DSP/FX on the channels. They cannot be used with BASS_ChannelGetData or BASS_ChannelGetLevel to visualise the channels, for example, but that could be achieved via a DSP function instead.
; Related .......: BASS_MIDI_StreamCreate, BASS_MIDI_StreamCreateFile
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamEvents($handle, $mode, $events, $length = 0)
	Local $pEvents
	Select
		Case IsDllStruct($events)
			$pEvents = DllStructGetPtr($events)
		Case IsPtr($events)
			$pEvents = $events
		Case IsArray($events)
			Local $sStruct = ''
			For $i = 1 To $events[0][0]
				$sStruct &= 'DWORD;DWORD;DWORD;DWORD;DWORD;'
			Next
			Local $tStruct = DllStructCreate($sStruct)
			For $i = 1 To $events[0][0]
				DllStructSetData($tStruct, ($i - 1) * 5 + 1, $events[$i][0])
				DllStructSetData($tStruct, ($i - 1) * 5 + 2, $events[$i][1])
				DllStructSetData($tStruct, ($i - 1) * 5 + 3, $events[$i][2])
				DllStructSetData($tStruct, ($i - 1) * 5 + 4, $events[$i][3])
				DllStructSetData($tStruct, ($i - 1) * 5 + 5, $events[$i][4])
			Next
			$pEvents = DllStructGetPtr($tStruct)
			$length = $events[0][0]
		Case Else
			Return SetError(1, 0, -1)
	EndSelect

	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamEvents", "DWORD", $handle, "DWORD", $mode, "PTR", $pEvents, "DWORD", $length)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = $BASS_DWORD_ERR Then Return SetError(_BASS_ErrorGetCode(), 0, -1)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamEvents


; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamGetChannel
; Description ...: Retrieves a MIDI channel's stream handle, which can be used to set DSP/FX on it.
; Syntax ........: _BASS_MIDI_StreamGetChannel($handle, $chan)
; Parameters ....: - $handle - The MIDI stream.
;                  - $chan - The MIDI channel... 0 = channel 1.
; Return values .: Success - If successful, the MIDI channel's stream handle is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: By default, MIDI channels do not have streams assigned to them; a MIDI channel only gets a stream when this function is called, which it then keeps until the MIDI stream is freed. MIDI channel streams can also be freed before then via BASS_StreamFree. Each MIDI channel stream increases the CPU usage slightly, even if there are no DSP/FX set on them, so for optimal performance they should not be activated when unnecessary.
;                  The MIDI channel streams have a different path to the final mix than the BASSMIDI reverb/chorus processing, which means that the reverb/chorus will not be present in the data received by any DSP/FX set on the streams and nor will the reverb/chorus be applied to the DSP/FX output; the reverb/chorus processing will use the channel's original data.
;                  The MIDI channel streams can only be used to set DSP/FX on the channels. They cannot be used with BASS_ChannelGetData or BASS_ChannelGetLevel to visualise the channels, for example, but that could be achieved via a DSP function instead.
; Related .......: BASS_MIDI_StreamCreate, BASS_MIDI_StreamCreateFile
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamGetChannel($handle, $chan)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamGetChannel", "DWORD", $handle, "DWORD", $chan)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamGetChannel



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamGetEvent
; Description ...: Retrieves the current value of an event in a MIDI stream channel.
; Syntax ........: _BASS_MIDI_StreamGetEvent($handle, $chan, $event)
; Parameters ....: - $handle - The MIDI stream to get the event value from.
;                  - $chan - The MIDI channel to get the event value from... 0 = channel 1.
;                  - $event - The event value to retrieve, which can be one of the following
; -------------------------------------------------------------------------------------------------------------------------------
; 						Event types, with param definitions.
; 							$MIDI_EVENT_NOTE			- Press or release a key.
; 															- param : LOBYTE = key number (0-127, 60=middle C), HIBYTE = velocity (0=release, 1-127=press).
; 							$MIDI_EVENT_PROGRAM			- Select the preset/instrument to use. Standard soundfont presets follow the General MIDI standard, and generally also include Roland GS variations in other banks (accessible via the MIDI_EVENT_BANK event).
; 															- param : preset number (0-127).
; 							$MIDI_EVENT_CHANPRES		- Set the channel pressure.
; 															- param : pressure level (0-127).
; 							$MIDI_EVENT_PITCH			- Set the pitch wheel position.
; 															- param : pitch wheel position (0-16383, 8192=normal/middle).
; 							$MIDI_EVENT_BANK			- Select the bank to use (MIDI controller 0).
; 															- param : bank number (0-127).
; 							$MIDI_EVENT_MODULATION		- Set the modulation (MIDI controller 1)
; 															- param : modulation level (0-127).
; 							$MIDI_EVENT_PORTATIME		- Set the portamento time (MIDI controller 5).
; 															- param : portamento time (0-127).
; 							$MIDI_EVENT_VOLUME			- Set the volume (MIDI controller 7).
; 															- param : volume level (0-127).
; 							$MIDI_EVENT_PAN				- Set the pan position (MIDI controller 10).
; 															- param : pan position (0-127, 64=middle).
; 							$MIDI_EVENT_EXPRESSION		- Set the expression (MIDI controller 11).
; 															- param : expression level (0-127).
; 							$MIDI_EVENT_SUSTAIN			- Set the sustain switch (MIDI controller 64).
; 															- param : enable sustain? (0-63=no, 64-127=yes).
; 							$MIDI_EVENT_PORTAMENTO		- Set the portamento switch (MIDI controller 65).
; 															- param : enable portamento? (0-63=no, 64-127=yes).
; 							$MIDI_EVENT_RESONANCE		- Set the low-pass filter resonance (MIDI controller 71, NRPN 121h)
; 															- param : resonance level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_RELEASE			- Set the release time (MIDI controller 72, NRPN 166h)
; 															- param : release time (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_ATTACK			- Set the attack time (MIDI controller 73, NRPN 163h)
; 															- param : attack time (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_CUTOFF			- Set the low-pass filter cutoff (MIDI controller 74, NRPN 120h)
; 															- param : cutoff level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_PORTANOTE		- Set the portamento start key; the next note starts at this key (MIDI controller 84).
; 															- param : key number (1-127, 60=middle C).
; 							$MIDI_EVENT_REVERB			- Set the reverb send level (MIDI controller 91)
; 															- param : reverb level (0-127).
; 							$MIDI_EVENT_CHORUS			- Set the chorus send level (MIDI controller 93)
; 															- param : chorus level (0-127).
; 							$MIDI_EVENT_SOUNDOFF		- Stop all sounds (MIDI controller 120).
; 															- param : not used.
; 							$MIDI_EVENT_RESET			- Reset controllers (MIDI controller 121), that is modulation=0, expression=127, sustain=0, portamento=0, release time=64, attack time=64, pitch wheel=8192, channel pressure=0.
; 															- param : not used.
; 							$MIDI_EVENT_NOTESOFF		- Release all keys (MIDI controller 123).
; 															- param : not used.
; 							$MIDI_EVENT_MODE			- Set poly/mono mode (MIDI controllers 126 & 127).
; 															- param : mode (0=poly, 1=mono).
; 							$MIDI_EVENT_PITCHRANGE		- Set pitch wheel range (MIDI RPN 0).
; 															- param : range in semitones.
; 							$MIDI_EVENT_FINETUNE		- Set the fine tuning (MIDI RPN 1).
; 															- param : finetune in cents (0-16383, 0=-100, 8192=normal, 16383=+100).
; 							$MIDI_EVENT_COARSETUNE		- Set the coarse tuning (MIDI RPN 2).
; 															- param : finetune in semitones (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUMS			- Set the percussion/drums channel switch.
; 															- param : use drums? (0=no, 1=yes).
; 							$MIDI_EVENT_DRUM_CUTOFF		- Set the low-pass filter cutoff of a drum key (MIDI NRPN 14knh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = cutoff level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUM_RESONANCE	- Set the low-pass filter resonance of a drum key (MIDI NRPN 15knh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = resonance level (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUM_COARSETUNE	- Set the coarse tuning of a drum key (MIDI NRPN 18knh).
; 															- param : LOBYTE = key number (0-127), HIBYTE = finetune in semitones (0-127, 0=-64, 64=normal, 127=+63).
; 							$MIDI_EVENT_DRUM_FINETUNE	- Set the fine tuning of a drum key (MIDI NRPN 19knh).
; 															- param : LOBYTE = key number (0-127), HIBYTE = finetune in cents (0-127, 0=-100, 64=normal, 127=+100).
; 							$MIDI_EVENT_DRUM_LEVEL		- Set the level of a drum key (MIDI NRPN 1Aknh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = level (0-127, 127=normal/full).
; 							$MIDI_EVENT_DRUM_PAN		- Set the pan position of a drum key (MIDI NRPN 1Cknh).
; 															- param : LOBYTE = key number (0-127), HIBYTE = pan position (0-127, 0=random, 64=middle).
; 							$MIDI_EVENT_DRUM_REVERB		- Set the reverb send level of a drum key (MIDI NRPN 1Dknh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = reverb level (0-127).
; 							$MIDI_EVENT_DRUM_CHORUS		- Set the chorus send level of a drum key (MIDI NRPN 1Eknh)
; 															- param : LOBYTE = key number (0-127), HIBYTE = chorus level (0-127).
; 						Global events.
; 							$MIDI_EVENT_TEMPO			- Set the tempo (MIDI meta event 81). Changing the tempo affects the stream length, and the BASS_ChannelGetLength byte value will no longer be valid.
; 															- param : tempo in microseconds per quarter note.
; 							$MIDI_EVENT_MASTERVOL		- Set the master volume.
; 															- param : volume level (0-16383, 0=silent, 16363=normal/full).
; 							$MIDI_EVENT_REVERB_TIME		- Set the reverb time.
; 															- param : reverb time in milliseconds.
; 							$MIDI_EVENT_REVERB_DELAY	- Set the reverb delay.
; 															- param : reverb delay in millisecond 10ths.
; 							$MIDI_EVENT_REVERB_LOCUTOFF	- Set the reverb low-pass cutoff
; 															- param : reverb low-pass cutoff in hertz (0=off).
; 							$MIDI_EVENT_REVERB_HICUTOFF	- Set the reverb high-pass cutoff
; 															- param : reverb high-pass cutoff in hertz (0=off).
; 							$MIDI_EVENT_REVERB_LEVEL	- Set the reverb level
; 															- param : reverb level (0=off, 100=0dB, 200=+6dB).
; 							$MIDI_EVENT_CHORUS_DELAY	- Set the chorus delay.
; 															- param : chorus delay in millisecond 10ths.
; 							$MIDI_EVENT_CHORUS_DEPTH	- Set the chorus depth.
; 															- param : chorus depth in millisecond 10ths.
; 							$MIDI_EVENT_CHORUS_RATE		- Set the chorus rate.
; 															- param : chorus rate in hertz 100ths.
; 							$MIDI_EVENT_CHORUS_FEEDBACK	- Set the chorus feedback level.
; 															- param : chorus feedback level (0=-100%, 100=off, 200=+100%).
; 							$MIDI_EVENT_CHORUS_LEVEL	- Set the chorus level.
; 															- param : chorus level (0=off, 100=0dB, 200=+6dB).
; 							$MIDI_EVENT_CHORUS_REVERB	- Set the chorus send to reverb level.
; 															- param : chorus send to reverb level (0=off, 100=0dB, 200=+6dB).
; 						Other (non-MIDI) events.
; 							$MIDI_EVENT_MIXLEVEL		- Set the level.
; 															- param : the level (0=silent, 100=0dB, 200=+6dB).
; 							$MIDI_EVENT_TRANSPOSE		- Transpose all notes. Changes take effect from the next note played, and affect melodic channels only (not drum channels).
; 															- param : transposition amount in semitones (0=-100, 100=normal, 200=+100).
; -------------------------------------------------------------------------------------------------------------------------------
; Return values .: Success - If successful, the event value is returned
;                  Failure - Returns -1 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Syncs can be used to be informed of when event values change.
; Related .......: BASS_MIDI_StreamEvent
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamGetEvent($handle, $chan, $event)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamGetEvent", "DWORD", $handle, "DWORD", $chan, "DWORD", $event)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = $BASS_DWORD_ERR Then Return SetError(_BASS_ErrorGetCode(), 0, -1)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamGetEvent



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamGetEvents
; Description ...: Retrieves the events in a MIDI file stream.
; Syntax ........: _BASS_MIDI_StreamGetEvents($handle, $track, $filter = 0)
; Parameters ....: - $handle - The MIDI stream to get the events from.
;                  - $track - The track to get the events from... 0 = 1st track.
;                  - $filter - The type of events to retrieve... 0 = all events. See BASS_MIDI_StreamEvent for a list of possible event types.
;                  - $returntype - if not 1 then a struct instead of an array is returned: $aEvents[0][0] = number of events in struct; $aEvents[0][1] = Events-struct
; Return values .: Success - Returns array of events
;                              $aEvents[0][0] = Number of events in array
;                              $aEvents[1][0] = First event type.
;                              $aEvents[1][1] = First event parameter.
;                              $aEvents[1][2] = MIDI channel
;                              $aEvents[1][3] = The position of the event, in ticks.
;                              $aEvents[1][4] = The position of the event, in bytes.
;                              ...
;                              $aEvents[n][0] = Nth event type.
;                              $aEvents[n][1] = Nth event parameter.
;                              $aEvents[n][2] = MIDI channel
;                              $aEvents[n][3] = The position of the event, in ticks.
;                              $aEvents[n][4] = The position of the event, in bytes.
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......:
; Related .......: BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamGetEvent, BASS_MIDI_EVENT structure
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamGetEvents($handle, $track, $filter = 0, $returntype = 1)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "dword", "BASS_MIDI_StreamGetEvents", "dword", $handle, "dword", $track, "dword", $filter, "ptr", 0)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = $BASS_DWORD_ERR Then Return SetError(_BASS_ErrorGetCode(), 0, 0)

	Local $iCnt = $BASS_ret_[0]

	Switch $returntype
		Case 1
			Local $aEvents[$iCnt + 1][5]
			$aEvents[0][0] = $iCnt
		Case Else
			Local $aEvents[1][2]
			$aEvents[0][0] = $iCnt
	EndSwitch

	If $iCnt < 1 Then Return $aEvents

	Local $sStruct = ''
	For $i = 1 To $iCnt
		$sStruct &= 'DWORD;DWORD;DWORD;DWORD;DWORD;'
	Next
	Local $tStruct = DllStructCreate($sStruct)

	$BASS_ret_ = DllCall($_ghBassMidiDll, "dword", "BASS_MIDI_StreamGetEvents", "dword", $handle, "dword", $track, "dword", $filter, "ptr", DllStructGetPtr($tStruct))
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = $BASS_DWORD_ERR Then Return SetError(_BASS_ErrorGetCode(), 0, 0)

	Switch $returntype
		Case 1
			For $i = 1 To $iCnt
				$aEvents[$i][0] = DllStructGetData($tStruct, ($i - 1) * 5 + 1)
				$aEvents[$i][1] = DllStructGetData($tStruct, ($i - 1) * 5 + 2)
				$aEvents[$i][2] = DllStructGetData($tStruct, ($i - 1) * 5 + 3)
				$aEvents[$i][3] = DllStructGetData($tStruct, ($i - 1) * 5 + 4)
				$aEvents[$i][4] = DllStructGetData($tStruct, ($i - 1) * 5 + 5)
			Next
		Case Else
			$aEvents[0][1] = $tStruct
	EndSwitch

	Return $aEvents
EndFunc   ;==>_BASS_MIDI_StreamGetEvents



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamGetFonts
; Description ...: Retrieves the soundfont configuration of a MIDI stream, or the default soundfont configuration.
; Syntax ........: _BASS_MIDI_StreamGetFonts($handle)
; Parameters ....: - $handle - The MIDI stream to retrieve the soundfont configuration of... 0 = get default soundfont configuration.
; Return values .: Success - Returns array of soundfonts
;                              $aFonts[0][0] = number of soundfonts in array
;                              $aFonts[1][0] = first Soundfont handle
;                              $aFonts[1][1] = Preset number of first SF
;                              $aFonts[1][2] = Base bank number of first SF
;                              ...
;                              $aFonts[n][0] = nth Soundfont handle
;                              $aFonts[n][1] = Preset number of nth SF
;                              $aFonts[n][2] = Base bank number of nth SF
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: When a soundfont matching the MIDI file is loaded, it will be the first element in the returned configuration.
; Related .......: BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamSetFonts, BASS_MIDI_FONT structure
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamGetFonts($handle)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamGetFonts", "DWORD", $handle, "PTR", 0, "DWORD", 0)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = $BASS_DWORD_ERR Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Local $iCount = $BASS_ret_[0]
	Local $aFonts[$iCount + 1][3]
	$aFonts[0][0] = $iCount
	If $iCount < 1 Then Return $aFonts
	Local $sStruct = ''
	For $i = 1 To $iCount
		$sStruct &= "dword;int;int;"
	Next
	Local $tStruct = DllStructCreate($sStruct)
	$BASS_ret_ = DllCall($_ghBassMidiDll, "DWORD", "BASS_MIDI_StreamGetFonts", "DWORD", $handle, "PTR", DllStructGetPtr($tStruct), "DWORD", $iCount)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = $BASS_DWORD_ERR Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	For $i = 1 To $iCount
		$aFonts[$i][0] = DllStructGetData($tStruct, ($i - 1) * 3 + 1)
		$aFonts[$i][1] = DllStructGetData($tStruct, ($i - 1) * 3 + 2)
		$aFonts[$i][2] = DllStructGetData($tStruct, ($i - 1) * 3 + 3)
	Next
	Return $aFonts
EndFunc   ;==>_BASS_MIDI_StreamGetFonts



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamGetMark
; Description ...: Retrieves a marker from a MIDI stream.
; Syntax ........: _BASS_MIDI_StreamGetMark($handle, $type, $index = -1)
; Parameters ....: - $handle - The MIDI stream to retrieve the marker from.
;                  - $type - The type of marker to retrieve, one of the following. BASS_MIDI_MARK_CUECue events (MIDI meta event 7).
;                            BASS_MIDI_MARK_KEYSIG Key signature events (MIDI meta event 89). The marker text is in the form of "a b", where a is the number of sharps (if positive) or flats (if negative), and b signifies major (if 0) or minor (if 1).
;                            BASS_MIDI_MARK_LYRIC Lyric events (MIDI meta event 5).
;                            BASS_MIDI_MARK_MARKER Marker events (MIDI meta event 6).
;                            BASS_MIDI_MARK_TEXT Text events (MIDI meta event 1).
;                            BASS_MIDI_MARK_TIMESIG Time signature events (MIDI meta event 88). The marker text is in the form of "a/b c d", where a is the numerator, b is the denominator, c is the metronome pulse, and d is the number of 32nd notes per MIDI quarter-note.
;                  - $index - The marker to retrieve... 0 = the first, -1 = all.
; Return values .: Success - Returns array of marker information
;                            $aMark[0][0] = number of markers in array
;                            $aMark[1][0] = The MIDI track containing the marker... 0 = first.
;                            $aMark[1][1] = The position (in bytes) of the marker.
;                            $aMark[1][2] = The marker text.
;                            ...
;                            $aMark[n][0] = The MIDI track containing the marker... 0 = first.
;                            $aMark[n][1] = The position (in bytes) of the marker.
;                            $aMark[n][2] = The marker text.
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: The markers are ordered chronologically.
;                  Syncs can be used to be informed of when markers are encountered during playback.
;                  If a lyric marker text begins with a / (slash) character, that means a new line should be started. If the text begins with a \ (backslash) character, the display should be cleared. Lyrics can sometimes be found in BASS_MIDI_MARK_TEXT instead of BASS_MIDI_MARK_LYRIC markers.
; Related .......: BASS_MIDI_StreamCreateFile, BASS_MIDI_MARK structure
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamGetMark($handle, $type, $index = -1)
	Local $BASS_ret_, $BASS_err_, $tStruct = DllStructCreate("DWORD;DWORD;PTR;")
	Local $aMark[1][3], $iCount = 0, $pStruct = DllStructGetPtr($tStruct)
	$aMark[0][0] = 0
	Switch $index
		Case -1 ;all
			While 1
				$BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_StreamGetMark", "DWORD", $handle, "DWORD", $type, "DWORD", $iCount, "PTR", $pStruct)
				If @error Then Return SetError(1, 1, 0)
				Switch $BASS_ret_[0]
					Case 0
						$BASS_err_ = _BASS_ErrorGetCode()
						Switch $BASS_err_
							Case $BASS_ERROR_ILLPARAM ;index not valid = no more markers
								Return $aMark
							Case Else
								Return SetError($BASS_err_, 0, 0)
						EndSwitch
					Case Else
						$iCount += 1
						ReDim $aMark[$iCount + 1][3]
						$aMark[$iCount][0] = DllStructGetData($tStruct, 1)
						$aMark[$iCount][1] = DllStructGetData($tStruct, 2)
						$aMark[$iCount][2] = DllStructGetData($tStruct, 3)
						$aMark[0][0] = $iCount
				EndSwitch
			WEnd
		Case Else
			$BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_StreamGetMark", "DWORD", $handle, "DWORD", $type, "DWORD", $index, "PTR", $pStruct)
			If @error Then Return SetError(1, 1, 0)
			Switch $BASS_ret_[0]
				Case 0
					$BASS_err_ = _BASS_ErrorGetCode()
					Switch $BASS_err_
						Case $BASS_ERROR_ILLPARAM ;index not valid = 0 markers
							Return $aMark
						Case Else
							Return SetError($BASS_err_, 0, 0)
					EndSwitch
				Case Else
					$iCount += 1
					ReDim $aMark[$iCount + 1][3]
					$aMark[$iCount][0] = DllStructGetData($tStruct, 1)
					$aMark[$iCount][1] = DllStructGetData($tStruct, 2)
					$aMark[$iCount][2] = DllStructGetData($tStruct, 3)
					$aMark[0][0] = $iCount
			EndSwitch
	EndSwitch
	Return $aMark
EndFunc   ;==>_BASS_MIDI_StreamGetMark



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamLoadSamples
; Description ...: Preloads the samples required by a MIDI file stream.
; Syntax ........: _BASS_MIDI_StreamLoadSamples($handle)
; Parameters ....: - $handle - The MIDI stream handle.
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Samples are normally loaded as they are needed while rendering a MIDI stream, which can result in CPU spikes, particularly with packed soundfonts. That generally won't cause any problems, but when smooth/constant performance is critical this function can be used to preload the samples before rendering, so avoiding the need to load them while rendering.
;                  Preloaded samples can be compacted/unloaded just like any other samples, so it is probably wise to disable the BASS_CONFIG_MIDI_COMPACT option when preloading samples, to avoid any chance of the samples subsequently being automatically unloaded.
;                  This function should not be used while the MIDI stream is being rendered, as that could interrupt the rendering.
; Related .......: BASS_MIDI_FontLoad, BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamSetFonts, BASS_CONFIG_MIDI_COMPACT
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamLoadSamples($handle)
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_StreamLoadSamples", "DWORD", $handle)
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamLoadSamples



; #FUNCTION# ====================================================================================================================
; Name ..........: _BASS_MIDI_StreamSetFonts
; Description ...: Applies a soundfont configuration to a MIDI stream, or sets the default soundfont configuration.
; Syntax ........: _BASS_MIDI_StreamSetFonts($handle, $aFonts)
; Parameters ....: - $handle - The MIDI stream to apply the soundfonts to... 0 = set default soundfont configuration.
;                  - $aFonts - The array of soundfonts to apply.
;                              $aFonts[0][0] = number of soundfonts in array
;                              $aFonts[1][0] = first Soundfont handle, previously initialized with BASS_MIDI_FontInit.
;                              $aFonts[1][1] = Preset number of first SF... -1 = use all presets in the soundfont.
;                              $aFonts[1][2] = Base bank number of first SF, or the bank number of the individual preset.
;                              ...
;                              $aFonts[n][0] = nth Soundfont handle, previously initialized with BASS_MIDI_FontInit.
;                              $aFonts[n][1] = Preset number of nth SF... -1 = use all presets in the soundfont.
;                              $aFonts[n][2] = Base bank number of nth SF, or the bank number of the individual preset.
; Return values .: Success - If successful, TRUE is returned
;                  Failure - Returns 0 and sets @ERROR to error returned by _BASS_ErrorGetCode()
; Author ........: Eukalyptus
; Modified ......:
; Remarks .......: Multiple soundfonts can be stacked, each providing different presets, for example. When a preset is present in multiple soundfonts, the earlier soundfont in the array has priority. When a soundfont matching the MIDI file is loaded, that remains loaded when calling this function, and has priority over all other soundfonts. When a preset is not available on a non-0 bank in any soundfont, BASSMIDI will try to fall back to bank 0.
;                  Changing the default configuration only affects subsequently created MIDI streams. Existing streams that are using the previous default configuration will continue to use that previous configuration.
; Related .......: BASS_MIDI_FontInit, BASS_MIDI_StreamCreateFile, BASS_MIDI_StreamGetFonts, BASS_MIDI_StreamLoadSamples, BASS_MIDI_FONT structure, BASS_CONFIG_MIDI_AUTOFONT, BASS_CONFIG_MIDI_DEFFONT
; Link ..........:
; Example .......:
; ===============================================================================================================================
Func _BASS_MIDI_StreamSetFonts($handle, $aFonts)
	If Not IsArray($aFonts) Or UBound($aFonts, 0) <> 2 Or UBound($aFonts, 2) <> 3 Then Return SetError(1, 1, False)
	If Not IsNumber($aFonts[0][0]) Or $aFonts[0][0] >= UBound($aFonts, 1) Then Return SetError(1, 2, False)
	Local $sStruct = ''
	For $i = 1 To $aFonts[0][0]
		$sStruct &= "dword;int;int;"
	Next
	Local $tStruct = DllStructCreate($sStruct)
	For $i = 1 To $aFonts[0][0]
		DllStructSetData($tStruct, ($i - 1) * 3 + 1, $aFonts[$i][0])
		DllStructSetData($tStruct, ($i - 1) * 3 + 2, $aFonts[$i][1])
		DllStructSetData($tStruct, ($i - 1) * 3 + 3, $aFonts[$i][2])
	Next
	Local $BASS_ret_ = DllCall($_ghBassMidiDll, "BOOL", "BASS_MIDI_StreamSetFonts", "DWORD", $handle, "PTR", DllStructGetPtr($tStruct), "DWORD", $aFonts[0][0])
	If @error Then Return SetError(1, 1, 0)
	If $BASS_ret_[0] = 0 Then Return SetError(_BASS_ErrorGetCode(), 0, 0)
	Return $BASS_ret_[0]
EndFunc   ;==>_BASS_MIDI_StreamSetFonts